#include<stack>
#include "KdTree.h"
#include "intersect_tri_cube.h"
#include "debug.h"
#include "Ray.h"
#include <float.h>

KdTree::KdTree(){
  min[X] = min[Y] = min[Z] = 1e99;
  max[X] = max[Y] = max[Z] = -1e99;
  shiny_min.setValue(1e99f, 1e99f, 1e99f);
  shiny_max.setValue(-1e99f, -1e99f, -1e99f);

  lists.push_back(TriangleList()); // 0th list signifies internal node 
  lists.push_back(TriangleList()); // 1st list is empty list
  lists.push_back(TriangleList()); // 2nd list is complete list

  nodes.push_back(node());  // 0th means no child (NODE_NULL)
  nodes.push_back(node());  // 1st is root (NODE_ROOT)
  rooti = NODE_ROOT;
  nodes[rooti].setLeaf();
  nodes[rooti].setListIdx(LIST_WHOLE);
  num_tris = 0;
  maxSideLength = 0.0;
  kdtreeDebug = 0;
  maxDepth = 8;
  maxItems = 10;
  rayId = 0;
  pthread_mutex_init(&lock, NULL);
  rays = 0;
  reordered = 0;
}

// recursive deletes nodes/children of nodes (but not triangles)
void KdTree::deleteNode(int ni){
  /*
  if (ni == 0)
    return;

  // delete children
  if (n->li == 0 && n->childi != 0){
    deleteNode(&(n->child[0]));  
    deleteNode(&(n->child[1]));
    if (!reordered)
      delete [] n->child;
  }

  // delete self
  //delete n;
  */
}
  
KdTree::~KdTree(){
  struct timeval start, end;
  cerr << "Deleting triangle tree..." << flush;
  gettimeofday(&start, NULL); // start delay measure

  // delete triangles
  cerr << "(triangles...)" << flush;
  TriangleList::deleteAll();

  cerr << "(octree nodes...)" << flush;
  // delete nodes
  //deleteNode(root);
//   if (reordered)
//     delete [] root;
//   else
//     delete root;

  // destroy mutex
  pthread_mutex_destroy( &lock );

  gettimeofday(&end, NULL);   // end delay measure
  cerr << " done. (" << delay(start, end) << "ms)" << endl;
}


// Resets tree, deleting all nodes
// This is called by rescanScene() to regenerate an octree
// because of changed geometry.
void KdTree::reset(){

  // clear tree visualization
  showCube(SbVec3f(0.0, 0.0, 0.0), SbVec3f(0.0, 0.0, 0.0),
	   SbVec3f(0.0, 0.0, 0.0), 1);

  // clear tree
  TriangleList::deleteAll();
  lists[LIST_WHOLE].clear();
  deleteNode(rooti);

  // reset variables
  nodes.clear();
  nodes.push_back(node()); // 0th means no children
  nodes.push_back(node()); // 1st is root
  rooti = NODE_ROOT;
  nodes[rooti].setLeaf();
  nodes[rooti].setListIdx(LIST_WHOLE);
  min[X] = min[Y] = min[Z] = 1e99;
  max[X] = max[Y] = max[Z] = -1e99;
  shiny_min.setValue(1e99f, 1e99f, 1e99f);
  shiny_max.setValue(-1e99f, -1e99f, -1e99f);
  maxSideLength = 0.0;
  num_tris = 0;
}
  
// set triangle tree parameters
void KdTree::setParams(int maxD, int maxI, int debug){
  maxDepth = maxD;
  maxItems = maxI;
  kdtreeDebug = debug;
}

// creates the octree structure
void KdTree::refine(int simpleDiv){
    
  // adjust first node boundaries
  
  // foreach triangle
  for (int ti=0; ti<lists[LIST_WHOLE].size(); ti++){

    Triangle *tri = lists[LIST_WHOLE].tri(ti);
    
    // foreach vertex
    for (int i=0; i<3; i++)
      
      // find max/min value
      for (int j=0; j<3; j++){
	
	if (tri->vert[i][j] > max[j]){
	  //showPoint((*I)->vert[i], 3.0); // debugging
	  max[j] = tri->vert[i][j];
	}
	if (tri->vert[i][j] < min[j]){
	  //showPoint((*I)->vert[i], 3.0); // debugging
	  min[j] = tri->vert[i][j];
	}

	// specular bounding box
	if (tri->isShiny()){
	  if (tri->vert[i][j] > shiny_max[j])
	    shiny_max[j] = tri->vert[i][j];
	  if (tri->vert[i][j] < shiny_min[j])
	    shiny_min[j] = tri->vert[i][j];
	}

      }
  }

  // setup boundaries of first node
  for (int i=0; i<3; i++){
    if (max[i]<min[i]){
      max[i] = 1;
      min[i] = -1;
    }
    if (max[i]-min[i] > maxSideLength)
      maxSideLength = max[i]-min[i];
    if (shiny_max[i]<shiny_min[i]){
      shiny_min[i] = -1;
      shiny_max[i] = 1;
    }
  }

  // find bounding sphere
  center = SbVec3f(min[X]+max[X], min[Y]+max[Y], min[Z]+max[Z])/2.0;
  SbVec3f maxCorner(max[X], max[Y], max[Z]);
  radius = (maxCorner-center).length();

  // shiny bounding sphere
  shiny_center = (shiny_min+shiny_max)/2.0;
  shiny_radius = (shiny_max-shiny_center).length();
  
#ifdef SHOWTRAV  
  // debugging
  nodes[rooti].min[0] = min[0];
  nodes[rooti].min[1] = min[1];
  nodes[rooti].min[2] = min[2];
  nodes[rooti].max[0] = max[0];
  nodes[rooti].max[1] = max[1];
  nodes[rooti].max[2] = max[2];
  nodes[rooti].depth = 0;
#endif
  
  if (min[X] == max[X] &&
      min[Y] == max[Y] &&
      min[Z] == max[Z]){
    cerr << "You forgot to set the size of the tree before refine().\n";
    exit(1);
  }

  // build list quickly
  sortAndRefine(rooti, 0, min, max);
  //refineNode(rooti, 0, simpleDiv, min, max);

  if (kdtreeDebug){
    showCube(SbVec3f(min[X], min[Y], min[Z]),
	     SbVec3f(max[X], max[Y], max[Z]),
	     SbColor(0.0, 1.0, 1.0));
  }

  compactNodes = &(nodes[0]);
}


const static int mod3[5] = { 0, 1, 2, 0, 1 };

// clips from list1 to list2
// using sutherland-hodge clipping
// a = axis
// p = planePoint[a]
// MinOrMax == 0: p is min
// MinOrMax == 1: p is max
template <int MinOrMax>
static void clip(const vector<SbVec3f> &list1, vector<SbVec3f> &list2,
		 int a, float p){
  
  for (int i=0; i<list1.size(); i++){      
      
    int j=(i+1)%list1.size();
    int l=0, r=0; // 0=out 1=in
    if (MinOrMax == 0){
      if (list1[i][a] >= p) l = 1;
      if (list1[j][a] >= p) r = 1;
    }
    else {
      if (list1[i][a] <= p) l = 1;
      if (list1[j][a] <= p) r = 1;
    }

    if (l^r){
      float t = (p-list1[i][a])/(list1[j][a]-list1[i][a]);
      SbVec3f isect = list1[i]+(list1[j]-list1[i])*t;
      isect[a] = p;
      if (list1[i][a]!=p && list1[j][a]!=p) list2.push_back(isect);
    }
    if (r) list2.push_back(list1[j]);

  }
}

static vector<SbVec3f> clipToVoxel(const float vmin[3],
				   const float vmax[3],
				   Triangle *t){

  vector<SbVec3f> listA, listB, *list1, *list2, *tmp;
  list1 = &listA;
  list2 = &listB;
  list2->push_back(t->vert[0]);
  list2->push_back(t->vert[1]);
  list2->push_back(t->vert[2]);
  
  // splits tri along each face,
  // returns set of verts as possible split points

  // loop over split planes (dims
  for (int a=0; a<3; a++){
    
    // clip with vmin
    if (list2->size()<3) return *list2;
    list1->clear();
    clip<0>(*list2, *list1, a, vmin[a]);
    
    // clip with vmax
    if (list1->size()<3) return *list1;
    list2->clear();
    clip<1>(*list1, *list2, a, vmax[a]);
  }
  
  return *list2;
}

inline int PerfectSplits(Triangle *t, const float vmin[3], const float vmax[3],
			  SbVec3f &pmin, SbVec3f &pmax){
  
  vector<SbVec3f> clipped = clipToVoxel(vmin, vmax, t);

  if (clipped.size() == 0)
    return 0;
  
  // only return extrema for each axis (bounding box)
  pmin=clipped[0], pmax=clipped[0];
  for (int i=1; i<clipped.size(); i++){
    for (int k=0; k<3; k++){
      if (clipped[i][k] > pmax[k]) pmax[k] = clipped[i][k];
      if (clipped[i][k] < pmin[k]) pmin[k] = clipped[i][k];
    }
  }

  return 1;
}


void KdTree::sortAndRefine(int ni, int depth, const float vmin[3],
			   const float vmax[3]){
  if (ni == NODE_NULL) return;
  const node &n = nodes[ni];
  
  // build event list
  vector<fullEvent_t> *eventList = new vector<fullEvent_t>;
  int numTris = lists[n.listIdx()].size();
  for (int ti=0; ti<lists[n.listIdx()].size(); ti++){
    int triId = lists[n.listIdx()].id(ti);
    Triangle *t = lists[n.listIdx()].tri(ti);
    SbVec3f pmin, pmax;
    int ret = PerfectSplits(t, vmin, vmax, pmin, pmax);
    if (ret == 0) cerr << "NO perfect splits!\n";
    for (int a=0; a<3; a++){
      if (pmin[a] == pmax[a])
	eventList->push_back(fullEvent_t(pmin[a], a, PLANAR, triId));
      else {
	eventList->push_back(fullEvent_t(pmin[a], a, BEGIN, triId)); 
	eventList->push_back(fullEvent_t(pmax[a], a, END, triId));
      }
    } // axis loop
  } // tri loop

  // sort event list
  radixSort(*eventList, eventList->size());

  // check vals
  bool valid=true;
  for (int i=0; i<eventList->size()-1; i++)
    if ((*eventList)[i].pos>(*eventList)[i+1].pos || ((*eventList)[i].pos==(*eventList)[i+1].pos && (*eventList)[i].eventType>(*eventList)[i+1].eventType)){
      valid=false;
      cerr << "Error at event="<<i<<": "
	   << (*eventList)[i].pos << " " << (*eventList)[i+1].pos << "\n";
    }
  if (valid) cerr << "Event list is sorted in ascending order.\n";
  else cerr << "ERROR: Event list is NOT sorted!\n";

  // refine
  TriangleList::allocBits();

  fastRefineNode(eventList, ni, numTris, depth, vmin, vmax);
}



const static double Ktrav = 0.3;
const static double Kint  = 0.8;
static inline float cost(float pl, float pr, float nl, float nr){
  const float l = 1.0f;//(nl&&nr) ? 1.0f : 0.8f;
  return l*(Ktrav+Kint*(pl*nl+pr*nr));
}

float KdTree::SAH(split_t p, const float vmin[3], const float vmax[3],
		  int nl, int nr, int np, int &leftOrRight){
  
  int k;
  float t, b, c, d, invD, pl, pr, el, er;
  const int a = p.axis;
  t = p.pos-vmin[a];
  k = mod3[a+1]; b = vmax[k]-vmin[k];
  k = mod3[a+2]; c = vmax[k]-vmin[k];
  d = vmax[a]-vmin[a];
  d = b*c + b*d + c*d; invD = 1.0f/d;
  pl = (         t *(b+c)+b*c)*invD;
  pr = ((vmax[a]-p.pos)*(b+c)+b*c)*invD;
	
  el = cost(pl, pr, nl+np, nr);
  er = cost(pl, pr, nl, nr+np);

  if (el<er){
    leftOrRight = LEFT;
    return el;
  }
  else {
    leftOrRight = RIGHT;
    return er;
  }
}

int KdTree::findSplit(int ni, float vmin[3], float vmax[3],
		      int &min_a, float &min_x,
		      TriangleList &leftList,
		      TriangleList &rightList){

  const node &n = nodes[ni];
  float min_c = Kint*lists[n.listIdx()].size();
  min_a = 0;
  min_x = (vmax[min_a]-vmin[min_a])*0.5f;
  int totalSplits = 0;
  TriangleList leftTris, rightTris, planarTris;
  
  // loop over tris
  for (int ti=0; ti<lists[n.listIdx()].size(); ti++){

    // clip tri to find splits
    Triangle *tri = lists[n.listIdx()].tri(ti);
    vector<SbVec3f> splits = clipToVoxel(vmin, vmax, tri);

    if (splits.size()<3) continue;
    totalSplits += splits.size();

    // loop over splits verts
    for (int s=0; s<splits.size(); s++){

      // loop over plane orientations
      for (int a=0; a<3; a++){

	// sort triangles into L, R, and planar lists
	leftTris.clear();
	rightTris.clear();
	planarTris.clear();

	// loop over triangles
	for (int tj=0; tj<lists[n.listIdx()].size(); tj++){

	  int triId = lists[n.listIdx()].id(tj);
	  Triangle *t = lists[n.listIdx()].tri(tj);
	  //Triangle *t = Triangle::globalTri(triId);
	  
	  // first clip triangle to voxel
	  vector<SbVec3f> result, clipped = clipToVoxel(vmin, vmax, t);

	  if (clipped.size()<3) continue; // is this necessary?

	  // test left voxel
	  clip<1>(clipped, result, a, splits[s][a]);

	  if (result.size()>2){
	    
	    // test for planar
	    int planar = 1;
	    for (int i=0; i<clipped.size(); i++)
	      if (clipped[i][a] != splits[s][a]){ planar = 0; break; }

	    if (planar) planarTris.add(triId);
	    else        leftTris.add(triId);
	  }

	  // test right voxel
	  result.clear();
	  clip<0>(clipped, result, a, splits[s][a]);

	  if (result.size()>2){
	    
	    // test for planar
	    int planar = 1;
	    for (int i=0; i<clipped.size(); i++)
	      if (clipped[i][a] != splits[s][a]){ planar = 0; break; }

	    if (!planar) rightTris.add(triId);
	  }

	}

        // SAH
	int leftOrRight;
	float c = SAH(split_t(splits[s][a], a), vmin, vmax, 
		      leftTris.size(), rightTris.size(), planarTris.size(),
		      leftOrRight);

	if (c < min_c){
	  
	  leftList.clear();  leftList.add(leftTris);
	  rightList.clear(); rightList.add(rightTris);
	  min_c = c;
	  min_x = splits[s][a];
	  min_a = a;
	  
	  if (leftOrRight == LEFT) leftList.add(planarTris);
	  else                     rightList.add(planarTris);

	}
	
      } // end axis loop
    } // end splits loop
  } // end triangle loop

  return(leftList.size() || rightList.size());
}


ostream& operator<< (ostream& os, const KdTree::fullEvent_t &e){

  os << "{pos=" << e.pos
     << " type=";
  switch (e.eventType){
  case KdTree::END:    os << "END";    break;
  case KdTree::PLANAR: os << "PLANAR"; break;
  case KdTree::BEGIN:  os << "BEGIN";  break;
  default:  os << e.eventType; break;
  }
  os << " axis=" << e.axis
     << " triId=" << e.triId << "}";
  
  return os;
}

void KdTree::printEventList(vector <fullEvent_t> *eventList){
  for (int i=0; i<eventList->size(); i++)
    cerr << "Event[" << i << "] = " << (*eventList)[i] << "\n";
}

void KdTree::printBitVector(vector<bool> *vec){
  for (int i=0; i<vec->size(); i++){
    cerr << ((*vec)[i] ? "1" : "0") << " ";
  }
  cerr << "\n";
}

int KdTree::findSplitFast(int ni, int niNumTris,
			  const float vmin[3], const float vmax[3],
			  vector <fullEvent_t> &eventList,
			  int &min_a, float &min_x,
			  int &min_leftOrRight,
			  int &min_left, int &min_right, int &min_planar){

  float nosplit_cost = Kint*niNumTris;
  float min_cost = nosplit_cost;
  min_leftOrRight = LEFT;
  
  int Nl[3], Np[3], Nr[3];
  for (int a=0; a<3; a++){
    Nl[a]=0;
    Np[a]=0;
    Nr[a]=niNumTris;
  }

  for (int i=0; i<eventList.size(); ){

    split_t p(eventList[i].pos, eventList[i].axis);
    int pBegin=0, pEnd=0, pPlanar=0;

    while (i<eventList.size() && eventList[i].axis == p.axis &&
	   eventList[i].pos == p.pos && eventList[i].eventType == END){
      pEnd++; i++;
    }

    while (i<eventList.size() && eventList[i].axis == p.axis &&
	   eventList[i].pos == p.pos && eventList[i].eventType == PLANAR){
      pPlanar++; i++;
    }

    while (i<eventList.size() && eventList[i].axis == p.axis &&
	   eventList[i].pos == p.pos && eventList[i].eventType == BEGIN){
      pBegin++; i++;
    }
    
    Np[p.axis] = pPlanar;
    Nr[p.axis] -= pPlanar;
    Nr[p.axis] -= pEnd;

    int leftOrRight;
    float cost;
    cost = SAH(p, vmin, vmax, Nl[p.axis], Nr[p.axis],
	       Np[p.axis], leftOrRight);

    if (cost<min_cost){
      min_cost = cost;
      min_a = p.axis;
      min_x = p.pos;
      min_leftOrRight = leftOrRight;
      min_left = Nl[p.axis];
      min_right = Nr[p.axis];
      min_planar = Np[p.axis];
    }

    Nl[p.axis] += pBegin;
    Nl[p.axis] += pPlanar;
    Np[p.axis] = 0;
  }

  return min_cost < nosplit_cost;
}

void KdTree::refineNode(int ni, int depth, int simpleDiv,
			float vmin[3], float vmax[3]){
  
  node *n = &nodes[ni];
  float nmin[3], nmax[3];

  // debugging info: show octree/child cell
  if (kdtreeDebug){
    SbColor color;
    color.setHSVValue(clamp((double)lists[n->listIdx()].size(),
			    (double)0.0,(double)36.0)/36.0,
		      1.0, 0.5);
    showCube(SbVec3f(vmin[X], vmin[Y], vmin[Z]),
	     SbVec3f(vmax[X], vmax[Y], vmax[Z]),
	     color);
  }	

  // subdivide node
  if (depth<maxDepth && n->isLeaf() && lists[n->listIdx()].size()>maxItems){
    int i;

    // debugging info
    if (kdtreeDebug){
      for (i=0; i<depth; i++)
	cerr << "  ";
      cerr << "refineNode(" << ni << ") "
	   << "(" << vmin[X] << ", " << vmax[X] << ") "
	   << "(" << vmin[Y] << ", " << vmax[Y] << ") "
	   << "(" << vmin[Z] << ", " << vmax[Z] << ")"
	   << " depth=" << depth
	   << " items=" << lists[n->listIdx()].size() << endl;;
    }

    int axis;
    TriangleList leftList, rightList;

    // set new size
    float newMid;
    if (simpleDiv){        // midpoint of longest axis
      float size[3];
      size[X] = vmax[X]-vmin[X];
      size[Y] = vmax[Y]-vmin[Y];
      size[Z] = vmax[Z]-vmin[Z];
      if (size[X]>size[Y] && size[X]>size[Z])
	axis = X;
      else if (size[Y]>size[Z])
	axis = Y;
      else
	axis = Z;
      newMid = 0.5*(vmax[axis]-vmin[axis])+vmin[axis];

      // need to set leftList and rightList
      // for (; I != lists[n->listInx()].end(); I++)
      //   if (intersect_tri_cube(boxcenter, boxhalfsize, **I))
      //     n->child[i].tri_list->add(*I);
    }
    else {                 // optimized point for subdivision
      int split = findSplit(ni, vmin, vmax, axis, newMid,
			    leftList, rightList);
      if (!split){
	if (kdtreeDebug) cerr << "no split found\n";
	goto refineDontSubdivide;
      }
      if (kdtreeDebug)
	cerr << "split found at a=" << axis << " p=" << newMid
	     << " left=" << leftList.size() << " right=" << rightList.size() << "\n";
    }

    // remove Triangles from tri_list, set terminal to 0
    if (ni != rooti && n->listIdx() != LIST_WHOLE)
      lists[n->listIdx()].clear();

    // update node
    n->setInner();
    n->setAxis(axis);
    n->setMid(newMid);
    
    // create children
    n->setChildIdx(nodes.size());
    nodes.push_back(node());
    nodes.push_back(node());
    n = &nodes[ni]; // n may have moved
    
    // for each child
    for (i=0; i<2; i++){

      // split box along axis
      nmin[axis] = (i==0) ? vmin[axis] : n->mid(); 
      nmax[axis] = (i==0) ? n->mid() : vmax[axis];

      // copy other box coords
      int oaxis = (axis+1)%3;
      nmin[oaxis] = vmin[oaxis];
      nmax[oaxis] = vmax[oaxis];
      oaxis = (axis+2)%3;
      nmin[oaxis] = vmin[oaxis];
      nmax[oaxis] = vmax[oaxis];

      // add appropriate children to new child node
      //   run triangle cube intersection for each triangle
      float boxcenter[3], boxhalfsize[3];
      boxcenter[0] = (nmax[X]+nmin[X])/2.0;
      boxcenter[1] = (nmax[Y]+nmin[Y])/2.0;
      boxcenter[2] = (nmax[Z]+nmin[Z])/2.0;
      boxhalfsize[0] = nmax[X]-boxcenter[0];
      boxhalfsize[1] = nmax[Y]-boxcenter[1];
      boxhalfsize[2] = nmax[Z]-boxcenter[2];

      // test each triangle
      lists.push_back(TriangleList());
      nodes[n->childIdx()+i].setLeaf();
      nodes[n->childIdx()+i].setListIdx(lists.size()-1);
#ifdef SHOWTRAV	    
      // debugging
      nodes[n->childIdx()+i].min = SbVec3f(nmin[X], nmin[Y], nmin[Z]);
      nodes[n->childIdx()+i].max = SbVec3f(nmax[X], nmax[Y], nmax[Z]);
      n->depth=depth+1;
#endif	    

      // add list to new child's list
      lists[nodes[n->childIdx()+i].listIdx()].add( i ? rightList : leftList);

      // refine new child node node
      refineNode(n->childIdx()+i, depth+1, simpleDiv, nmin, nmax);
      n = &nodes[ni]; // n may have moved
	
    } // end for(1..2)
      
  } // end if (depth<max_depth...)
    
  else{ // if (n->tri_list.size()){

    refineDontSubdivide:

    if (n->isLeaf())
      lists[n->listIdx()].resizeToFit(); // save memory

    // debugging info: print cell info
    if (kdtreeDebug){
      if (1 || lists[n->listIdx()].size()){ // all or just empty cells
	for (int x=0; x<depth; x++)
	  cerr << "  ";
	cerr << ni << " ";
	cerr << lists[n->listIdx()].size() << " triangles depth=" << depth << endl;
      }
    }
      
  }

} // end of refineNode();




void KdTree::fastRefineNode(vector<fullEvent_t> *eventList, int ni,
			    int niNumTris, int depth, 
			    const float vmin[3], const float vmax[3]){

  node *n = &nodes[ni];
  float nmin[3], nmax[3];

  // debugging info: show octree/child cell
  if (kdtreeDebug){
    SbColor color;
    color.setHSVValue(clamp((double)niNumTris,
			    (double)0.0,(double)36.0)/36.0,
		      1.0, 0.5);
    showCube(SbVec3f(vmin[X], vmin[Y], vmin[Z]),
	     SbVec3f(vmax[X], vmax[Y], vmax[Z]),
	     color);

    for (int i=0; i<depth; i++)
      cerr << "  ";
    cerr << "fastRefineNode(" << ni << ") "
	 << "(" << vmin[X] << ", " << vmax[X] << ") "
	 << "(" << vmin[Y] << ", " << vmax[Y] << ") "
	 << "(" << vmin[Z] << ", " << vmax[Z] << ")"
	 << " depth=" << depth
	 << " items=" << niNumTris
	 << " events=" << eventList->size()
	 << (eventList->size()>niNumTris*6?" (too many!)":" (fine)")
	 << endl;
    
  }	
  
  // subdivide node
  if (depth<maxDepth && niNumTris>maxItems){
    int i;

    int axis;

    // set new size
    float newMid;

    int leftOrRight, nL, nR, nP;
    int split = findSplitFast(ni, niNumTris, vmin, vmax, *eventList,
			      axis, newMid, leftOrRight,
			      nL, nR, nP);
    if (!split){
      if (kdtreeDebug){
	for (i=0; i<depth; i++)
	  cerr << "  ";
	cerr << "no split found\n";
      }
      goto dontSubdivide;
    }
    if (kdtreeDebug){
      for (i=0; i<depth; i++)
	cerr << "  ";
      cerr << "split found at a=" << axis << " p=" << newMid
	   << " left=" << nL << " right=" << nR << " planar=" << nP
	   << (leftOrRight==LEFT?" left":" right")
	   << "\n";
    }

    // build leftList, rightList
    //const int numTris = lists[LIST_WHOLE].size();
    //vector<bool> sides(numTris*2);

    // Step 1) Classification.
    // assign both to all tris
    //sides.assign(numTris*2, true);        // true on L and R
    TriangleList::setAllBits(true);
    //TriangleList::printBits();
      
    for (int i=0; i<eventList->size(); i++){
      if ((*eventList)[i].eventType == END &&
	  (*eventList)[i].axis == axis &&
	  (*eventList)[i].pos <= newMid)
	//sides[(*eventList)[i].triId*2+RIGHT] = false;
	TriangleList::setBit((*eventList)[i].triId, RIGHT, false);
      else if ((*eventList)[i].eventType == BEGIN &&
	       (*eventList)[i].axis == axis &&
	       (*eventList)[i].pos >= newMid)
	//sides[(*eventList)[i].triId*2+LEFT] = false;
	TriangleList::setBit((*eventList)[i].triId, LEFT, false);
      else if ((*eventList)[i].eventType == PLANAR &&
	       (*eventList)[i].axis == axis) {
	if ((*eventList)[i].pos < newMid ||
	    ((*eventList)[i].pos == newMid && leftOrRight == LEFT))
	  //sides[(*eventList)[i].triId*2+RIGHT] = false;
	  TriangleList::setBit((*eventList)[i].triId, RIGHT, false);
	if ((*eventList)[i].pos > newMid ||
	    ((*eventList)[i].pos == newMid && leftOrRight == RIGHT))
	  //sides[(*eventList)[i].triId*2+LEFT] = false; // R
	  TriangleList::setBit((*eventList)[i].triId, LEFT, false);
      }
    }


    // Step 2) Splicing E into E_LO and E_RO

    // Todo:
    // count lo and ro 1st to save memory
    // treat sides in an array fashion so no repeat code 
      
    // remove triangle on both sides but make a list of them
    vector<fullEvent_t> listLO, listRO, listBL, listBR;
    vector<int> tListBoth;
    int numL=0, numR=0;
    for (int i=0; i<eventList->size(); i++){
      int id = (*eventList)[i].triId;
      bool l = TriangleList::getBit(id, LEFT); //sides[id*2+LEFT];
      bool r = TriangleList::getBit(id, RIGHT); //sides[id*2+RIGHT];

      // foreach triangle
      if (((*eventList)[i].eventType == BEGIN ||
	   (*eventList)[i].eventType == PLANAR)
	   && (*eventList)[i].axis==0){
	if (l) numL++;
	if (r) numR++;
	if (l && r)
	  tListBoth.push_back(id);
      }

      if (l && r)
	1; // do nothing
      else if (l)
	listLO.push_back((*eventList)[i]);
      else if (r)
	listRO.push_back((*eventList)[i]);
      else 
	cerr << "Event is neither L or R!\n";	  
    }

    // resize
    delete eventList;
    vector<fullEvent_t>(listLO).swap(listLO);
    vector<fullEvent_t>(listRO).swap(listRO);

    // create new bounding boxes
    float mmax[3], mmin[3];
    mmax[0] = vmax[0]; mmin[0] = vmin[0];
    mmax[1] = vmax[1]; mmin[1] = vmin[1];
    mmax[2] = vmax[2]; mmin[2] = vmin[2];
    mmax[axis] = newMid;
    mmin[axis] = newMid;
    
    // loop over common triangles
    for (int i=0; i<tListBoth.size(); i++){

      int id = tListBoth[i];

      // clip i against everything to compute new events

      Triangle *t = TriangleList::globalTri(id);
      SbVec3f pmin, pmax;

      // left
      int ret = PerfectSplits(t, vmin, mmax, pmin, pmax);
      if (ret == 0) cerr << "NO perfect splits!\n";
	
      for (int a=0; a<3; a++){
	if (pmin[a] == pmax[a])
	  listBL.push_back(fullEvent_t(pmin[a], a, PLANAR, id));
	else {
	  listBL.push_back(fullEvent_t(pmin[a], a, BEGIN, id)); 
	  listBL.push_back(fullEvent_t(pmax[a], a, END, id));
	}
      } // axis loop

	// right
      ret = PerfectSplits(t, mmin, vmax, pmin, pmax);
      if (ret == 0) cerr << "NO perfect splits!\n";
	
      for (int a=0; a<3; a++){
	if (pmin[a] == pmax[a])
	  listBR.push_back(fullEvent_t(pmin[a], a, PLANAR, id));
	else {
	  listBR.push_back(fullEvent_t(pmin[a], a, BEGIN, id)); 
	  listBR.push_back(fullEvent_t(pmax[a], a, END, id));
	}
      } // axis loop
    }

    // sort event list
    radixSort(listBL, listBL.size());
    radixSort(listBR, listBR.size());

    // Step 4) Merge 4 strains.
    vector<fullEvent_t> *newL, *newR;
    newL = new vector<fullEvent_t>;
    newR = new vector<fullEvent_t>;

    vector<fullEvent_t>::iterator iLO, iRO, iBL, iBR;

    // populate left 
    iLO = listLO.begin();
    iBL = listBL.begin();
    while(iLO != listLO.end() || iBL != listBL.end()){
      if (iLO == listLO.end() ||
	  iBL != listBL.end() &&
	  (iBL->pos < iLO->pos ||
	   (iBL->pos == iLO->pos &&
	    iBL->eventType < iLO->eventType)))
	newL->push_back(*iBL++);
      else
	newL->push_back(*iLO++);
    }

    // populate right
    iRO = listRO.begin();
    iBR = listBR.begin();
    while(iRO != listRO.end() || iBR != listBR.end()){
      if (iRO == listRO.end() ||
	  iBR != listBR.end() &&
	  (iBR->pos < iRO->pos ||
	   (iBR->pos == iRO->pos &&
	    iBR->eventType < iRO->eventType)))
	newR->push_back(*iBR++);
      else
	newR->push_back(*iRO++);
    }

    // recurse
    
    // update node
    n->setInner();
    n->setAxis(axis);
    n->setMid(newMid);

    // create children
    n->setChildIdx(nodes.size());
    nodes.push_back(node());
    nodes.push_back(node());
    n = &nodes[ni]; // n may have moved
    
    fastRefineNode(newL, n->childIdx()+0, numL, depth+1, vmin, mmax);
    n = &nodes[ni]; // n may have moved
    fastRefineNode(newR, n->childIdx()+1, numR, depth+1, mmin, vmax);
    
    return;
  }

  dontSubdivide:

  if (ni == NODE_ROOT){
      delete eventList;
  }
  else {
    
    // create empty triangle list
    lists.push_back(TriangleList());
    n->setLeaf();
    n->setListIdx(lists.size()-1);

    // populate triangle list from eventList
    for (int i=0; i<eventList->size(); i++){
      if (((*eventList)[i].eventType == PLANAR ||
	  (*eventList)[i].eventType == BEGIN) && (*eventList)[i].axis==0){
	int id = (*eventList)[i].triId;
	lists[n->listIdx()].add(id);
      }
    }

    delete eventList;
    lists[n->listIdx()].resizeToFit(); // save memory
  }

  // debugging info: print cell info
  if (kdtreeDebug){
    if (1 || lists[n->listIdx()].size()){ // all or just empty cells
      for (int x=0; x<depth; x++)
	cerr << "  ";
      cerr << "done with ni=" << ni
	   << " tris=" << lists[n->listIdx()].size()
	   << " depth=" << depth << endl;
    }
  }
  
} // end of fastRefineNode();


//vector<KdTree::stackElem_t> KdTree::kdstack;

KdTree::stackElem2_t KdTree::stk[64];

int KdTree::intersect_ray(Ray &r) {

  // count rays
  r.ctx->rays++;
  
  SbVec3f invDir = SbVec3f(1.0f/r.dir[X], 1.0f/r.dir[Y], 1.0f/r.dir[Z]);
  
//   if (invDir[X]<0.0f) { front[X]=1; back[X]=0; }
//   else                { front[X]=0; back[X]=1; }
//   if (invDir[Y]<0.0f) { front[Y]=1; back[Y]=0; }
//   else                { front[Y]=0; back[Y]=1; }
//   if (invDir[Z]<0.0f) { front[Z]=1; back[Z]=0; }
//   else                { front[Z]=0; back[Z]=1; }
//   const int *back  = frontBacks[a^7];
//   double t0 = (min[axis]-r.pos[axis])*invDir[axis];
//   double t1 = (max[axis]-r.pos[axis])*invDir[axis];
  
//   if (invDir[axis]<0.0f){
//     const double t = t0;
//     t0 = t1;
//     t1 = t;
//   }

  static const int frontBacks[8][3] = {
    { 0, 0, 0 },
    { 1, 0, 0 },
    { 0, 1, 0 },
    { 1, 1, 0 },
    { 0, 0, 1 },
    { 1, 0, 1 },
    { 0, 1, 1 },
    { 1, 1, 1 }
  };
  static const int frontBacks8[8][3] = {
    { 0, 0, 0 },
    { 8, 0, 0 },
    { 0, 8, 0 },
    { 8, 8, 0 },
    { 0, 0, 8 },
    { 8, 0, 8 },
    { 0, 8, 8 },
    { 8, 8, 8 }
  };

  unsigned int a = 0;
  a |= ((*(unsigned int *)&r.dir[X]) & 1<<31) >> 31;
  a |= ((*(unsigned int *)&r.dir[Y]) & 1<<31) >> 30;
  a |= ((*(unsigned int *)&r.dir[Z]) & 1<<31) >> 29;
  const int *front = frontBacks[a];
  const int *front8 = frontBacks8[a];
  const int axis = compactNodes[rooti].axis();

  double ts[3][2] = {
   { (min[0]-(double)r.pos[0])*invDir[0],
     (max[0]-(double)r.pos[0])*invDir[0] },
   { (min[1]-(double)r.pos[1])*invDir[1],
     (max[1]-(double)r.pos[1])*invDir[1] },
   { (min[2]-(double)r.pos[2])*invDir[2],
     (max[2]-(double)r.pos[2])*invDir[2] }
  };

  double t0 = MAX3(ts[0][front[0]],
		   ts[1][front[1]],
		   ts[2][front[2]]);
		   
  double t1 = MIN3(ts[0][front[0]^1],
		   ts[1][front[1]^1],
		   ts[2][front[2]^1]);

  //const double eps = 0.000001;
  //if (t0 < eps) t0 = eps;
  if (t0 < EPSILON_OI) t0 = EPSILON_OI;
  if (t0 > t1) return 0;
  
#ifdef SHOWTRAV  
  // debugging
  if (DEBUG_STATE){
    cerr << "Size: " << n->max-n->min << "\n";
    SbColor color;
    color.setHSVValue(n->depth/8.0, 1.0, 0.5);
    showCube(n->min, n->max, color);
  }
#endif
  
  int ni = rooti;
  stackElem2_t *stkElem = stk;
  double tm;
  
//#define EPSILON_IR 0.001

  const char *base = (char *)compactNodes;
  node *n = &compactNodes[ni];
  
  while (1) {

    while (!NODE_ISLEAF(n)){

      // this code is for axis-aligned direction handling (special case)
      // see paper Sec.3.3 in Efficient Parametric Algorithm for Octree Traversal
      const unsigned int axis = n->axis();
//       if (isinf(t1) == 1){
// 	if (r.pos[axis] < n->mid) tm =  t1;
// 	else                      tm = -t1;
//       }
//       else
	tm = (n->mid() - (double)r.pos[axis])*invDir[axis];

      if (tm < t0)      // cull front side
	n = (node *)(base+NODE_CHILDOFFSET(n) + (front8[axis]^8));
      else if (tm > t1) // cull back side
	n = (node *)(base+NODE_CHILDOFFSET(n) + front8[axis]);
      else {
	stkElem->n = (node *)(base+NODE_CHILDOFFSET(n)+(front8[axis]^8));
	stkElem->t0 = tm;
	stkElem->t1 = t1;

	stkElem++;
	
	n = (node *)(base+NODE_CHILDOFFSET(n) + front8[axis]);
	t1 = tm;
      }
    }

    // leaf
//     const double t0eps = 1.0-FLT_EPSILON;
//     const double t1eps = 1.0+FLT_EPSILON;
    if (lists[n->listIdx()].size() &&
	//lists[n->listIdx()].intersect_ray(r, t0*t0eps, t1*t1eps)) return 1;
	lists[n->listIdx()].intersect_ray(r, t0-EPSILON_OI, t1+EPSILON_OI)) return 1;
    
    if (stkElem-- == stk) return 0;
    n = stkElem->n;
    t0 = stkElem->t0;
    t1 = stkElem->t1;
  }
}

int KdTree::proc_subtree(const double &t0, const double &t1,
			       const int ni,
			       const treeParam &tp) const {
} // end proc_subtree()

  // single sided intersection routine
  //   calls regular intersection routine
int KdTree::ss_intersect_ray(Ray &r){
  int retVal = intersect_ray(r);
    
  if (retVal && r.tri->getNormal().dot(r.dir)>0.0){
    Ray new_r = Ray(r.pos+(r.t+0.0001)*r.dir, r.dir, r.ctx);
    retVal = ss_intersect_ray(new_r);
    r.t += new_r.t;
  }

  return(retVal);
}

typedef struct statData {
  int tris;
  int nodes;
  int leafs;
  int emptyLeafs;
  int leafDepthTotal;
  int nodesDepthLimit;
  int minLeafDepth;
  int maxLeafDepth;
  int maxLeafTris;
} statData_t;

void KdTree::printStats(int ni, int depth){

  node *n = &nodes[ni];
  static statData_t stat;

  // if no parameters, call self on root
  if (ni == NODE_NULL){
    stat.tris            = 0;
    stat.nodes           = 0;
    stat.leafs           = 0;
    stat.emptyLeafs      = 0;
    stat.leafDepthTotal  = 0;
    stat.nodesDepthLimit = 0;
    stat.minLeafDepth    = 99999;
    stat.maxLeafDepth    = 0;
    stat.maxLeafTris     = 0;

    printStats(rooti, 0);  // root is node (leaf?) at depth=0, it's children are d=1
    
    cerr << "KdTree statistics:\n"
         << "tris                   : " << num_tris << "\n"
	 << "tris in tree           : " << stat.tris
	 << " (" << stat.tris*100.0f/num_tris << "%)\n"
         << "leaf nodes             : " << stat.leafs << " (empty: " << stat.emptyLeafs
	 << " = " << stat.emptyLeafs*100.0f/stat.leafs << "%)\n"
	 << "avg tris/non-empty leaf: "
	 << stat.tris/(float)(stat.leafs-stat.emptyLeafs) << " ("
	 << "max: " << stat.maxLeafTris << ")\n"
	 << "nodes                  : " << stat.nodes
	 << " (inner: " << stat.nodes-stat.leafs << " = "
	 << (stat.nodes-stat.leafs)*100.0f/stat.nodes << "%)\n"
         << "leaf min/avg/max depth : "
	 << stat.minLeafDepth << "/"
	 << stat.leafDepthTotal/(float)stat.leafs << "/"
	 << stat.maxLeafDepth << "\n"
	 << "leaves at depth lim    : " << stat.nodesDepthLimit << " ("
	 << stat.nodesDepthLimit*100.0f/stat.leafs << "%)\n"
	 << "";
    return;
  }

  stat.nodes++;

  // leaf
  if (n->isLeaf()){
    
    stat.leafs++;
    stat.tris += lists[n->listIdx()].size();
    stat.leafDepthTotal += depth;

    if (depth==maxDepth)
      stat.nodesDepthLimit++;

    if (lists[n->listIdx()].size() > stat.maxLeafTris)
      stat.maxLeafTris = lists[n->listIdx()].size();
    
    if (lists[n->listIdx()].size() == 0)
      stat.emptyLeafs++;

    if (depth<stat.minLeafDepth)
      stat.minLeafDepth = depth;

    if (depth>stat.maxLeafDepth)
      stat.maxLeafDepth = depth;
  }

  // otherwise, recurse over children
  else {
    printStats(n->childIdx()+0, depth+1);
    printStats(n->childIdx()+1, depth+1);
  }

  return;
}


void KdTree::reorderNodes(){

  // count nodes
  node *p, *q;
  int nodeCnt = nodes.size();

  // allocate space for nodes
  int nodeIdx = 0;
  compactNodes = new node[nodeCnt];
  compactNodes[nodeIdx++] = nodes[NODE_NULL];
  compactNodes[nodeIdx++] = nodes[NODE_ROOT];
  root = &compactNodes[NODE_ROOT];
  rooti = NODE_ROOT;
//  root->setChildIdx(NODE_ROOT);

  p = root;
  q = &compactNodes[nodeCnt];

  while (p != q){
    if (!p->isLeaf()){
      if (nodeIdx>=nodeCnt) {
	cerr << "WTF too many nodes\n";
	break;
      }

      compactNodes[nodeIdx+0] = nodes[p->childIdx()+0];
      compactNodes[nodeIdx+1] = nodes[p->childIdx()+1];

      p->setChildIdx( nodeIdx );
      nodeIdx+=2;
    }
    p++;
  }

  cerr << "Moved " << nodeCnt << " nodes.\n";
  nodes.clear();

//   // make relative offsets
//   p = root; int pi=root;
//   q = &compactNodes[nodeCnt];
//   while (p != q){
//     if (!p.isLeaf()){
//       int idx = p.childIdx();
//       p.setChildIdx(idx-pi);
//     }
//     p++; pi++;
//   }
  
  reordered=1;
  return;
}


// finds sign of floating point number:
//   1 (negative): flips all bits
//   0 (positive): flips sign
inline unsigned int FloatFlip(unsigned int f){
  unsigned int mask = -(int)(f>>31) | 0x80000000;
  return f ^ mask;
}
inline unsigned int FloatFlipF(unsigned int &f){
  unsigned int mask = -(int)(f>>31) | 0x80000000;
  f ^= mask;
}

// flips a float back
inline unsigned int FloatFlipI(unsigned int f){
  unsigned int mask = ((f>>31)-1) | 0x80000000;
  return f ^ mask;
}

// expects unsigned int, and eventType
// picks off 10-bit chunks
#define _0(x,t,a) ( (x & 0x3F) << 4 | t << 2 | a)
#define _1(x,t,a) ( x >>  6 & 0x3FF )
#define _2(x,t,a) ( x >> 16 & 0x3FF )
#define _3(x,t,a) ( x >> 26 )

// radix sort by Michael Herf
// http://www.stereopsis.com/radix.html
// email to use: stereopsis att mherf dt com 
void KdTree::radixSort(vector<fullEvent_t> &orig, int numVals){
  unsigned int i;
  vector<fullEvent_t> sort(numVals);

  // sort[i] => sort[i].posui
  
  const unsigned int kHist = 1024;
  unsigned int b0[kHist * 4];
  unsigned int *b1 = b0+kHist;
  unsigned int *b2 = b1+kHist;
  unsigned int *b3 = b2+kHist;

  for (i=0; i<kHist*4; i++)
    b0[i] = 0;

  // 1) parallel histogramming pass
  for (i=0; i<numVals; i++){
    unsigned int f = FloatFlip(orig[i].posui);
    b0[_0(f, orig[i].eventType, orig[i].axis)]++;
    b1[_1(f, orig[i].eventType, orig[i].axis)]++;
    b2[_2(f, orig[i].eventType, orig[i].axis)]++;
    b3[_3(f, orig[i].eventType, orig[i].axis)]++;
  }

  // 2) sum histograms: each histogram entry records the
  // number of values preceding itself.
  unsigned int sum0=0, sum1=0, sum2=0, sum3=0, tsum;
  for (i=0; i<kHist; i++){
    tsum = b0[i]+sum0;
    b0[i] = sum0-1;
    sum0 = tsum;

    tsum = b1[i]+sum1;
    b1[i] = sum1-1;
    sum1 = tsum;

    tsum = b2[i]+sum2;
    b2[i] = sum2-1;
    sum2 = tsum;

    tsum = b3[i]+sum3;
    b3[i] = sum3-1;
    sum3 = tsum;
  }

  // byte 0
  // floatflip entire value, read/write hist, write out flipped
  for (i=0; i<numVals; i++){
    unsigned int f = orig[i].posui;
    FloatFlipF(f);
    unsigned int pos = _0(f, orig[i].eventType, orig[i].axis);
    orig[i].posui = f;
    sort[++b0[pos]] = orig[i];
  }

  // byte 1
  // read/write hist, copy
  for (i=0; i<numVals; i++){
    unsigned int s = sort[i].posui;
    unsigned int pos = _1(s, sort[i].eventType, sort[i].axis);
    orig[++b1[pos]] = sort[i];
  }

  // byte 2
  // read/write hist, copy
  for (i=0; i<numVals; i++){
    unsigned int s = orig[i].posui;
    unsigned int pos = _2(s, orig[i].eventType, orig[i].axis);
    sort[++b2[pos]] = orig[i];
  }
  
  // byte 3
  // read/write hist, copy & flip out
  for (int i=0; i<numVals; i++){
    unsigned int s = sort[i].posui;
    unsigned int pos = _3(s, sort[i].eventType, sort[i].axis);
    sort[i].posui = FloatFlipI(s);
    orig[++b3[pos]] = sort[i];
  }

}
