/////////////////////////////////////////////////////////////////////////////
//                                                                           
//  WWWWW                                                                     
//  WW  WW                                                                    
//  WW  WW                                                                    
//  WWWWW    WWWW   WWWW    WWW                                               
//  WW          WW  WW WW  WW WW                                              
//  WW       WWWWW  WW WW  WWWWW                                              
//  WW      WW  WW  WW WW  WW                                                 
//  WW       WWWWW  WW WW   WWWW                                              
//                                                                           
//                                                                           
//
// File: Main.cxx
// Description: Main ray tracer routine (start up)
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Main routine for the Pane ray tracer.
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// Usage: pane [OPTION]... [IV FILE]
//
// Synopsis: pane is a ray tracer.
//           Input is a scene in the Open Inventor format (scene.iv)
//           Output is an image (scene.png).
//
// $Id: Main.cxx 1186 2007-11-12 09:02:06Z beason $
//
////////////////////////////////////////////////////////////////////////////

#include "Main.h"
#include "debug.h"
#include "Ray.h"
#include "Texture.h"
#include "Triangle.h"
#include "Octree.h"
#include "KdTree.h"
#include "GlobalVars.h"
#include "Pane.h"
#include "sobol.h"
#include "CameraList.h"
#include "Misc.h"
#include "PreviewQuad.h"
#include <signal.h>

//namespace paneMain {

int DEBUG_STATE=0;                         // debugging mode if true
int STOP_RENDERING=0;                      // stop rendering if true
const int mutexNum = 16;                   // max number of threads
pthread_mutex_t mutexes[mutexNum];         // mutual exclusion locks for threads
SoSeparator *scnRoot;                      // overall scene graph root
RNG mydrand48;

// vector finite
int finite(SbVec3f &c){
  return( finite(c[X]) && finite(c[Y]) && finite(c[Z]) );
}

// vector mult
SbVec3f mul(SbVec3f &x, SbVec3f &y, float c){
  return(SbVec3f(x[X]*y[X]*c,
		 x[Y]*y[Y]*c,
		 x[Z]*y[Z]*c));
}

//''WW'WW''''''''WW''WWW'''''''''''''''''''''''WWW''''''''''''''
// WW            WW WW  W                     WW  W             
//WWWW WW WWW   WWW  WW    WWW  WW  WWW   WW  WW     WWW WWWW W 
// WW  WW WW W W WW   WW  WW W WWWW WW W WWWW WW    W WW WW WW W
// WW  WW WW W W WW W  WW WW   WW   WW W WW   WW  W W WW WW WW W
// WW  WW WW W  WWW  WWW   WWW  WWW WW W  WWW  WWW   WWW WW WW W
//                                                              
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Finds the first SoCamera in a scene
//
void findSceneCam(GlobalVars *gv){
  // Search for camera in scene graph
  SoSearchAction sa;
  sa.setType(SoCamera::getClassTypeId());
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.apply(gv->root);
  if (sa.getPath())
    gv->cam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
}

// forward declaration
void animateCamera(void *data, SoSensor *);


// threads assistance
namespace {
  typedef void (Pane::*memberTypeNoArg)();
  typedef void (Pane::*memberTypeIntArg)(int);
  
  typedef struct memberFunc {
    Pane *obj;
    memberTypeNoArg member1;
    memberTypeIntArg member2;
  } memberFunc;
  
  void *runPaneMember(void *arg){
    memberFunc *mf = (memberFunc*) arg;
    if (mf->member1)
      ((mf->obj)->*(mf->member1))();
    else if (mf->member2)
      ((mf->obj)->*(mf->member2))(0);
  }
  
  pthread_t threads[8];
  int currThread=0;

  void callPaneMember(GlobalVars *gv,
		      memberTypeNoArg member1,
		      memberTypeIntArg member2){
    static memberFunc mf; // NOT thread safe, only for singular use
    mf.obj = gv->pane;
    mf.member1 = member1;
    mf.member2 = member2;
    pthread_create(&threads[currThread++], NULL, runPaneMember, &mf);
  }
}



// toggles the scene graph between two values
// previous(-2)/none(-1)/scene(0)/preview(1)
// see togglePreviewDisplay to turn on preview
void toggleSceneSwitch(GlobalVars *gv, int val1=-2, int val2=-2){
  static int oldval=-1;
  int curval = gv->sceneSwitch->whichChild.getValue();
  int newval;

  // equal vals: set specific val
  // diff vals: choose other  
  if (val1==val2) newval = val1; 
  else newval = (curval!=val1) ? val1 : val2;
  if (newval==-2) newval = oldval;

  if (newval>=gv->sceneSwitch->getNumChildren()) return;
  gv->sceneSwitch->whichChild.setValue(newval);
  if (curval!=oldval && curval!=newval) oldval = curval;
}



//WW''''''''''''''WWWWW'''''''''''''''''WW''WW''W'''''''''''''WW'WWW'''''''
//WW              WW                    WW  WW  W             WW  WW       
//WW W   WW  W WW WWWW  WW W  WW  WWW  WWWW WW  W  WWW WWW   WWW  WW   WW  
//WWW   WWWW W WW WW    WW W WWWW WW W  WW  WWWWW W WW WW W W WW  WW  WWWW 
//WW W  WW   W WW WW    WW W WW   WW W  WW  WW  W W WW WW W W WW  WW  WW   
//WW  W  WWW  WWW WWWWW  WW   WWW WW W  WW  WW  W  WWW WW W  WWW WWWW  WWW 
//             WW                                                          
//'''''''''''WWW'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Processes viewer keypresses
//
void keyEventHandler(void *gvp, SoEventCallback *callBack){
  const SoEvent *event = callBack->getEvent();
  GlobalVars *gv = (GlobalVars *)gvp;

  // display keypress letter
  if (!gv->figureMode){

    gv->letterSep->removeAllChildren();
    
    SoTransform *tran = new SoTransform();
    tran->translation.setValue(.8, .8, 0);
    tran->scaleFactor.setValue(.1, .1, 0);
    gv->letterSep->addChild(tran);
    
    SoMaterial *cubeMat = new SoMaterial();
    cubeMat->diffuseColor.setValue(0,0,0);
    cubeMat->shininess.setValue(0);
    cubeMat->ambientColor.setValue(0,0,0);
    cubeMat->emissiveColor.setValue(.6,.6,.6);
    gv->letterSep->addChild(cubeMat);
    
    SoCube *cube = new SoCube();
    gv->letterSep->addChild(cube);

    SoTransform *tran2 = new SoTransform();
    tran2->translation.setValue(-.8,-.6, .001);
    gv->letterSep->addChild(tran2);
    
    SoTransform *textTran = new SoTransform();
    textTran->scaleFactor.setValue(.08, .08, 0.01);
    gv->letterSep->addChild(textTran);
    
    SoMaterial *textMat = new SoMaterial();
    textMat->diffuseColor.setValue(0,0,0);
    textMat->shininess.setValue(0);
    textMat->ambientColor.setValue(0,0,0);
    //textMat->emissiveColor.setValue(0,1,0);
    gv->letterSep->addChild(textMat);

    SoFont *font = new SoFont();
    font->size.setValue(24);
    gv->letterSep->addChild(font);
    
    //SoText3 *text = new SoText3();
    SoText2 *text = new SoText2();
    string s = "[";
    char c = ((SoKeyboardEvent *)event)->getPrintableCharacter();
    if (!c) c = ' ';
    s += c;
    s += "]";
    text->string.setValue(s.c_str());
    //gv->letterSep->addChild(text);   // causes div-by-0 in SbMatrix.c++:733
  }

  // reorder nodes
  static int shiftPressed = 0;
  if (SO_KEY_PRESS_EVENT(event, LEFT_SHIFT)){
    shiftPressed = 1;
  }
  else if (SO_KEY_RELEASE_EVENT(event, LEFT_SHIFT)){
    shiftPressed = 0;
  }
  
  // reorder nodes
  if (SO_KEY_PRESS_EVENT(event, E)){
    gv->pane->rayAccel->reorderNodes();
  }
  
  // Render image when 'R' is pressed
  if (SO_KEY_PRESS_EVENT(event, R)){
    if (!gv->renderQuit){
      SbVec2s newSize(gv->imgDim[X]+60, gv->imgDim[Y]+32);
      SbVec2s actSize(gv->imgDim[X]/2, gv->imgDim[Y]/2);
      gv->xtViewer->setSize(newSize);
      //gv->xtViewer->setViewportRegion(actSize);
      gv->preview->togglePreviewDisplay(gv, true);
    }
    callPaneMember(gv, NULL, &Pane::render);
  }

  // Write rendered image to file when 'W' is pressed
  if (SO_KEY_PRESS_EVENT(event, W)){
    gv->pane->writeImg();
  }

  // Toggle headlight when 'H' is pressed
  if (SO_KEY_PRESS_EVENT(event, H)){
    int i = gv->headLightPos;
    i = (i+1)%3;
    gv->headLightPos = i;
    cerr << "Headlight ";
    switch(i){
    case HL_OFF:
      cerr << "OFF";
      gv->sceneHasLights--;
      break;
    case HL_CENTER:
      cerr << "CENTER";
      gv->sceneHasLights++;
      break;
    case HL_OFFCENTER:
      cerr << "OFFCENTER";
    }
    cerr << endl;
  }

  // turn up headlight
  if (SO_KEY_PRESS_EVENT(event, EQUAL)){ // up
    double i = gv->pane->headLight->intensity.getValue();
    cerr << "Increasing headLight intensity from "
	 << setiosflags(ios::showpoint | ios::fixed) << setprecision(8)
	 << i;
    resetios(cerr);
    i = pow(10.0, log10(i)+0.25);
    cerr << " to " << i << endl;
    gv->pane->headLight->intensity.setValue(i);
  }

  // turn down headlight
  if (SO_KEY_PRESS_EVENT(event, MINUS)){ // down
    double i = gv->pane->headLight->intensity.getValue();
    cerr << "Decreasing headLight intensity from "
	 << setiosflags(ios::showpoint | ios::fixed) << setprecision(8)
	 << i;
    resetios(cerr);
    i = pow(10.0, log10(i)-0.25);
    cerr << " to " << i << endl;
    gv->pane->headLight->intensity.setValue(i);
  }

  // Print camera info
  if (SO_KEY_PRESS_EVENT(event, C)){

    cerr << "Camera info:" << endl;
    SoOutput *out = new SoOutput();
    out->setFilePointer(stderr);
    SoWriteAction *wa = new SoWriteAction(out);

    // Search for camera in scene graph
    SoSearchAction sa;
    sa.setType(SoCamera::getClassTypeId());
    sa.setSearchingAll(FALSE); // don't look under off switches
    sa.apply(gv->root);
    if (sa.getPath())
      gv->cam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();

    if (shiftPressed){
      wa->apply(gv->preview->getCamera());
      const SbViewportRegion vp =
	gv->preview->getCamera()->getViewportBounds(gv->xtViewer->getViewportRegion());
      SbVec2s s = vp.getViewportSizePixels();
      cerr << "Pixel size: " << s[0] << "x" << s[1] << "\n";
    }
    else if (gv->cam->isOfType(SoPerspectiveCamera::getClassTypeId()))
      wa->apply((SoPerspectiveCamera*)gv->cam);
    else if (gv->cam->isOfType(SoOrthographicCamera::getClassTypeId()))
      wa->apply((SoOrthographicCamera*)gv->cam);
    else
      wa->apply(gv->cam);
  }

  // Save scene to file
  if (SO_KEY_PRESS_EVENT(event, S)){
    //cerr << "Camera info:" << endl;
    SoOutput *out = new SoOutput();
    //out->setFilePointer(stderr);
    const char *fn = "pane_dump.iv";
    out->openFile(fn);
    SoWriteAction *wa = new SoWriteAction(out);
    wa->apply(gv->root); //gv->cam);
    out->closeFile();
    cerr << "Dumped scene graph to " << fn << ".\n";
    delete out;
    delete wa;
  }
  
  // Align Camera to axes
  if (SO_KEY_PRESS_EVENT(event, Z)){
    static int rev=1;
    rev = !rev;

    findSceneCam(gv);
    gv->cam->orientation.setValue(SbVec3f(0, 1, 0), M_PI*rev);
  }
  if (SO_KEY_PRESS_EVENT(event, X)){
    static int orient = 1;
    orient *= -1;

    findSceneCam(gv);
    gv->cam->orientation.setValue(SbVec3f(0, 1, 0),
				  orient*1.57079632679489661922);
  }
  if (SO_KEY_PRESS_EVENT(event, Y)){
    static int orient = 1;
    orient *= -1;

    findSceneCam(gv);
    gv->cam->orientation.setValue(SbVec3f(1, 0, 0),
				  orient*1.57079632679489661922);
  }

  // position camera
  if (SO_KEY_PRESS_EVENT(event, V)){
    cerr << "Positioned camera" << endl;
    gv->cam->orientation.setValue(
      SbVec3f(-0.0014208, 0.999947, 0.0102222),  5.10545);
    gv->cam->position.setValue(-2298.35, 602.14, -1020.66);
  }

  // animate camera
  if (SO_KEY_PRESS_EVENT(event, A)){
    static SoTimerSensor *sensor;
    static int animating = 0;

    if (sensor == NULL)
      sensor = new SoTimerSensor(animateCamera, gv);
    
    // fixes default value (set to signify not to position camera)
    if (gv->timeVal<0.0)
      gv->timeVal = 0.0;

    if (animating){
      cerr << "Stopping animation.\n";
      sensor->unschedule();
      animating = 0;
    }
    else {
      cerr << "Animating.\n";
      sensor->setInterval(SbTime(0.015)); // 70 frames per second
      sensor->schedule();
      animating = 1;
    }
  }

  // animate camera
  if (SO_KEY_PRESS_EVENT(event, I)){
    gv->preview->togglePreviewDisplay(gv);
  }
  
  // rotate scene
  if (SO_KEY_PRESS_EVENT(event, O)){

    SbVec3f axis;
    float angle;
    gv->trans->rotation.getValue(axis, angle);
    gv->trans->rotation.setValue(axis, angle+M_PI/4.0);

    // recompute octree
    gv->pane->rescanScene();
  }

  // trace pixel
  if (SO_KEY_PRESS_EVENT(event, T)){
    gv->pane->render(1);
  }

  // show normals
  if (SO_KEY_PRESS_EVENT(event, N)){
    gv->showNormals = !gv->showNormals;
    if (gv->showNormals)
      cerr << "Showing normals.\n";
    else
      cerr << "Disabled showing normals.\n";
  }
  
  // shoot photons
  if (SO_KEY_PRESS_EVENT(event, P)){
    TIME(gv->pane->shootPhotons());
  }

  // debugging
  // irradiance estimate
  if (SO_KEY_PRESS_EVENT(event, D)){
    SbVec3f hit_x = SbVec3f(371.333801, 144.431351, 331.016296);
    SbVec3f hit_n = SbVec3f(-0.0148499319, -0.999874473, -0.00552253844);
    float irrad[3];

    int temp = DEBUG_STATE;
    DEBUG_STATE = 1;
    gv->pane->gPhotonMap->irradiance_estimate(
      irrad, hit_x.getValue(), (-hit_n).getValue(),
      gv->gPhotonEstimate);
    DEBUG_STATE = temp;    
    cerr << "irrad = "; printVec(SbVec3f(irrad)); cerr << endl;
  }

  // toggle geometry
  if (SO_KEY_PRESS_EVENT(event, G)){
    toggleSceneSwitch(gv, 0, -1);
  }

  // toggle crosshair
  if (SO_KEY_PRESS_EVENT(event, PERIOD)){

    int val = gv->crossSwitch->whichChild.getValue();
    int newVal = (val==0)?-1:0;
    gv->crossSwitch->whichChild.setValue(newVal);
  }
  
  // turns surfaces black, for photon viewing
  if (SO_KEY_PRESS_EVENT(event, B)){

    SoCallbackAction cb;
    cb.addPreCallback(SoMaterial::getClassTypeId(),
		      Pane::matBlackCBHandler, (void *)NULL);
    cb.apply(gv->scene);
  }

  // cycle window background color
  if (SO_KEY_PRESS_EVENT(event, K)){
    static int col = 0;
    // 0 = windowBG, 1 = black, 2 = gray, 3 = white, 4 = purple
    col = (col+1)%5;
    switch(col){
    case 0:
      gv->xtViewer->setBackgroundColor(gv->windowBG);
      break;
    case 1:
      gv->xtViewer->setBackgroundColor(SbColor(0.0, 0.0, 0.0));
      break;
    case 2:
      gv->xtViewer->setBackgroundColor(SbColor(0.25, 0.25, 0.25));
      break;
    case 3:
      gv->xtViewer->setBackgroundColor(SbColor(1.0, 1.0, 1.0));
      break;
    case 4:
      gv->xtViewer->setBackgroundColor(SbColor(0.5, 0.0, 0.5));
      break;
      
    default:
      break;
    }
  }
  
}

// Animate scene
void rotateScene(void *data, SoSensor *){
  GlobalVars *gv = (GlobalVars *)data;

  // rotate scene
  SbVec3f axis;
  float angle;
  gv->trans->rotation.getValue(axis, angle);
  angle += 0.02;
  if (angle>M_PI*2.0)
    angle = 0.0;
  gv->trans->rotation.setValue(axis, angle);

  // update octree if not set to do it automatically
  if (!gv->automaticUpdate)
    gv->pane->rescanScene();
}


//''''''''''WW'''''''''''''''WW''''''''WWW'''''''''''''''''''''''''''''
//                           WW       WW  W                            
// WWW WWW  WW WWWW W   WWW WWWW  WW  WW     WWW WWWW W   WW  W WW  WWW
//W WW WW W WW WW WW W W WW  WW  WWWW WW    W WW WW WW W WWWW WW   W WW
//W WW WW W WW WW WW W W WW  WW  WW   WW  W W WW WW WW W WW   WW   W WW
// WWW WW W WW WW WW W  WWW  WW   WWW  WWW   WWW WW WW W  WWW WW    WWW
//                                                                     
//'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Reposition the camera
//
void animateCamera(void *data, SoSensor *){
  GlobalVars *gv = (GlobalVars *)data;

  SoCamera *cam = (SoCamera *)gv->xtViewer->getCamera();

  if (gv->camList && cam->isOfType(SoPerspectiveCamera::getClassTypeId()) )
    gv->camList->positionCameraCubic((SoPerspectiveCamera*)cam,
				     (1.0-cos(gv->timeVal*M_PI))/2.0);
  
  // increment time

  //float secs = gv->frames * 0.15;
  gv->timeVal += 1.0/(gv->frames-1);
  if (gv->timeVal>1.0)
    gv->timeVal = 0.0;
}



//'''''''''''''''''''''''''''WWW''WW''''''''''''''''''''''''WWW''WWWW'
//                          WW  W WW                       WW  W WW  W
// WWWW  WWW  WW  WWW   WW  WW    WWW   WWW WWW   WWW  WW  WW    WWWW 
//WWW   WW W WWWW WW W WWWW WW    WW W W WW WW W W WW WWWW WW    WW  W
//  WWW WW   WW   WW W WW   WW  W WW W W WW WW W W WW WW   WW  W WW  W
//WWWW   WWW  WWW WW W  WWW  WWW  WW W  WWW WW W  WWW  WWW  WWW  WWWW 
//                                                 WW                 
//'''''''''''''''''''''''''''''''''''''''''''''''WWW''''''''''''''''''
//
// Callback to update octree if scene changes
//
void sceneChangeCB(void *data, SoSensor *sensor){
  ((GlobalVars *)data)->pane->rescanScene();
}




//'''''''''''''''''''''WW''''WW''''''WW'''''
//                     WWW  WWW             
//WWW  W WW  WW    WWW W WWW WW  WWW WW WWW 
//WW W WW   WW W  W WW W  W  WW W WW WW WW W
//WW W WW   WW W  W WW W     WW W WW WW WW W
//WWW  WW    WW    WWW W     WW  WWW WW WW W
//WW                WW                      
//WW''''''''''''''WWW'''''''''''''''''''''''
//
// Main subroutine for the ray tracer
//

int progMain(int argc, char* argv[]){

  setupSignals();
  enableExceptions();

  // store name of program for debugging
  getProgramName() = argv[0];

  // start delay measure
  struct timeval totalStart, totalEnd;
  gettimeofday(&totalStart, NULL); 
  
  // Global variables object
  GlobalVars gv;
  gv.readOpts(argc, argv);

  // Initialize Inventor
  if (gv.renderQuit){
    //SoDB::init();
    SoInteraction::init();
  }
  else {
    SoXt::init(argv[0]);     // works better than SoDB::init(), thanks Chuck.
  }
  SoMatPlus::initClass();
  
  Pane pane(&gv);
  gv.pane = &pane;

  // seed random number generator
  //srand48(time(NULL));             // random
  srand48(gv.srandSeed);             // deterministic (default 0)
  mydrand48.setSeed(gv.srandSeed);   // platform independent RNG
  sobinit(20);                       // initialize sobol point generator

  // determine number of processors
  if (gv.numThreads == -1){

    int procs = 0;
    char line[1000];
    ifstream cpuinfo("/proc/cpuinfo");
    if (cpuinfo){
      regex_t regex;
      int reti;
      reti = regcomp(&regex, "^processor", 0);
      if (reti)
	cerr << "Warning: Regular expression parsing failure, reading # cpus!"
	     << "Try using -jN to manually set number of threads.\n";
      else {
	while (cpuinfo){
	  cpuinfo.getline(line, 1000, '\n');
	  reti = regexec(&regex, line, 0, NULL, 0);
	  if (!reti)
	    procs++;
	}
      regfree(&regex);
      }
      cpuinfo.close();
    }
    gv.numThreads = MAX2(1, procs);
    cerr << "Using " << gv.numThreads << " threads.\n";
  }
      
  cerr << "Reading scene file...";
  struct timeval start, end;
  gettimeofday(&start, NULL); // start delay measure

  // Read scene
  if (!gv.inImgFile.empty()){
    cerr << "(null scene)";
    gv.scene = new SoSeparator();
  }
  else {
    
    // read scene data
    SoInput *sceneIn = new SoInput();
    if (!sceneIn->openFile(gv.sceneFile.c_str())){
      cerr << "Unable to open scene file: " << gv.sceneFile << endl;
      exit(1);
    }
    
    gv.scene = SoDB::readAll(sceneIn);
    delete sceneIn;
  }
  gv.scene->ref();
  if (!gv.scene){
    cerr << "Error: scene separator is null!" << endl;
    exit(1);
  }

  // replace File nodes of scene with editable children
  // FIXME UNDO NEXT TWO LINES
  //pane.makeEditableChildren(gv.scene);
  //pane.makeEditableChildren(gv.scene); // for SoFile nodes inside SoFile nodes
  
  // Does scene have a camera?
  SoSearchAction sa;
  sa.setType(SoCamera::getClassTypeId());
  sa.setSearchingAll(FALSE); // don't look under off switches
  sa.apply(gv.scene);
  SoCamera *sceneCam = NULL;
  SoFullPath *fullPath = NULL;  
  if (sa.getPath()){
    sceneCam = (SoCamera *)((SoFullPath *)sa.getPath())->getTail();
    fullPath = (SoFullPath *) sa.getPath();
  }
  
  // If camera exists, and is of a known type, move it to top of
  // scene root. If it's of unknown type, remove it and create
  // a perspective camera at scene root.
  //
  // We do this because the ExaminerViewer attaches to and modifies
  // the first camera (if it exists), modifying the scenegraph.
  //
  // Since we want to detect changes to the scenegraph, not including
  // view/camera changes, this is bad. Therefore we move the camera outside
  // of the scenegraph.
  if (sceneCam){

    // find parent node of camera
    SoGroup *parent = (SoGroup *)fullPath->getNodeFromTail(1);

    // identify camera type and if it's unknown,
    // create a new camara.
    if (sceneCam->isOfType(SoOrthographicCamera::getClassTypeId())){
      cerr << "Orthographic camera found...";
      gv.cam = sceneCam;
    }
    else if (sceneCam->isOfType(SoPerspectiveCamera::getClassTypeId())){
      cerr << "Perspective camera found...";
      gv.cam = sceneCam;
    }
    else {
      cerr << "Unknown camera type found, creating perspective camera...";

      // copy common stuff
      gv.cam = new SoPerspectiveCamera();
      gv.cam->viewportMapping = sceneCam->viewportMapping.getValue();
      gv.cam->position        = sceneCam->position.getValue();
      gv.cam->orientation     = sceneCam->orientation.getValue();
      gv.cam->aspectRatio     = sceneCam->aspectRatio.getValue();
      gv.cam->focalDistance   = sceneCam->focalDistance.getValue();
    }

    // remove the original camera
    // gv.cam->ref();
    // parent->removeChild(sceneCam);
    // gv.cam->unrefNoDelete();
  }

  // otherwise create a camera
  else { 
    gv.cam = new SoPerspectiveCamera();
    gv.cam->viewAll(gv.scene, SbViewportRegion());
  }


  //                   scnRoot
  //                (aka gv->root)
  //                       |
  //            -----------------------------------...
  //            |          |         |           
  //      crosshairSep  camera   sceneSwitch   
  //            |                    |--------------|
  //           ...                sceneSep      previewSep
  //                           (aka gv->scene)
  //                                 |
  //                         --------------... (rest of the scene)
  //                         |         
  //                     transform    
  
  // create scene root
  gv.root = new SoSeparator();
  gv.root->ref();
  scnRoot = gv.root;
  
  // add crosshair
  SoSeparator *chSep = new SoSeparator();
  SoMaterial *chMat = new SoMaterial();
  //chMat->emissiveColor.setValue(0.9, 0.6, 0.9);
  chMat->emissiveColor.setValue(0.2, 0.8, 0.2);
  chSep->addChild(chMat);
  SoTransform *chXform = new SoTransform();
  chXform->scaleFactor.setValue(0.003, 0.003, 0.003);
  chXform->translation.setValue(0.0, 0.0, -0.01);
  chSep->addChild(new SoResetTransform());
  chSep->addChild(chXform);
  chSep->addChild(new SoCube());

  // add letter display
  gv.letterSep = new SoSeparator();

  // put crosshair in switch so it can be toggled
  gv.crossSwitch = new SoSwitch();
  gv.crossSwitch->addChild(chSep);
  gv.crossSwitch->addChild(gv.letterSep);
  if (!gv.figureMode)
    gv.crossSwitch->whichChild.setValue(SO_SWITCH_ALL);
  gv.root->addChild(gv.crossSwitch);

  // add camera
  gv.root->addChild(gv.cam);

  // add scene after adding transform (for rotation) to it
  gv.trans = new SoTransform();
  gv.scene->insertChild(gv.trans, 0);

  // put scene in switch so it can be toggled
  gv.sceneSwitch = new SoSwitch();
  gv.sceneSwitch->addChild(gv.scene);
  gv.sceneSwitch->whichChild.setValue(0);
  
  gv.root->addChild(gv.sceneSwitch);
  gv.scene->unref(); // now that addChild ref'd, can unref earlier ref()
  gv.root->addChild(new SoResetTransform());

  // add preview
  if (!gv.renderQuit) {
    gv.preview = new PreviewQuad();
    gv.sceneSwitch->addChild(gv.preview->constructScene(gv.imgDim, gv.numThreads));
  }
    
  gettimeofday(&end, NULL);   // end delay measure
  cerr << "done. (" << delay(start, end) << "ms)" << endl;

  // compute normals
  if (gv.computeNormals){
    cerr << "\nComputing normals...\n";
    computeNormals(gv.scene);
  }
  
  cerr << "Initializing ray tracing engine...";
  gettimeofday(&start, NULL); // start delay measure
  int numTris = pane.init();  // initialize pane
  gettimeofday(&end, NULL);   // end delay measure
  cerr << numTris << " triangles read ("
       << delay(start, end) << "ms)" << endl;
  gv.times.octreeCreation = delay(start, end);

  // make camera list from file
  gv.camList = new CameraList();
  gv.camList->readList(string());

  // timeVal of <0 (the default) signifies not to set the camera position
  if (gv.cam->isOfType(SoPerspectiveCamera::getClassTypeId()) &&
      gv.timeVal>=0.0){
    gv.camList->positionCameraCubic( (SoPerspectiveCamera *)gv.cam,
				     gv.timeVal );
    gv.timeMin = gv.timeVal;
    gv.timeMax = gv.timeVal + 1.0f/(gv.frames-1);
    if (gv.timeMax > 1.0f) gv.timeMax = 1.0f;
  }

  // fixes default value (set to signify not to position camera)
  if (gv.timeVal<0.0)
    gv.timeVal = 0.0;
  
  // Create Examiner Viewer
  if (! gv.renderQuit){

    // This returns a main window to use.
    // If unsuccessful, exit.
    Widget myWindow = SoXt::init(argv[0]); // pass the app name
    if (myWindow == NULL) exit(1);

    // animate scene
    if (gv.animateScene){      
      //SoTimerSensor *sensor = new SoTimerSensor(rotateScene, &gv);
      SoTimerSensor *sensor = new SoTimerSensor(animateCamera, &gv);
      sensor->setInterval(SbTime(0.015)); // 70 frames per second
      sensor->schedule();
    }
    
    // Set up keyboard
    SoEventCallback *callBack = new SoEventCallback();
    callBack->addEventCallback(SoKeyboardEvent::getClassTypeId(),
			       keyEventHandler, &gv);
    gv.root->addChild(callBack);

    cerr <<
      "\n"
      "Press ESC, then R to render.\n"
      "Then press W to write image.\n"
      "\n"
      "Keys:\n"
  "  P  shoot photons                R  render scene\n"      
  "  H  toggle headlight             W  write image to disk\n"
  "\n"
      "More Keys:\n"
  "  +  increase headlight power     X  point camera along negative X axis\n"
  "  -  decrease headlight power     Y  point camera along negative Y axis\n"
  "  C  print camera node            Z  point camera along negative Z axis\n"
  "  S  save scene graph to file     O  rotate scene and recalculate octree\n"
  "  T  trace ray and show           G  toggle geometry\n"
  "  N  show normals for traced ray  A  animate camera using camlist.iv\n"
  "  .  Turn off crosshair           B  turn surfaces black\n"
  "  I  Toggle preview display       K  cycle window background color\n"
	 << endl;
    
    // Set up viewer
    SoXtExaminerViewer *myViewer = new SoXtExaminerViewer(myWindow);
    gv.xtViewer = myViewer;  // store viewer for use by animateCamera
   
    myViewer->setSceneGraph(gv.root);
    myViewer->setTitle("Pane Scene Viewer");
    if (!gv.fastTransparency){
      myViewer->setTransparencyType(SoGLRenderAction::DELAYED_BLEND);
      //myViewer->setTransparencyType(SoGLRenderAction::SORTED_OBJECT_BLEND);
    }
    //myViewer->setTransparencyType(SoGLRenderAction::SCREEN_DOOR);
    myViewer->setBackgroundColor(gv.windowBG);

    // FIGURE: remove me
    if (gv.figureMode)
      pane.shootPhotons();
    
    // Show scene
    myViewer->show();

    // automatically update octree on scene change
    if (gv.automaticUpdate){
      SoNodeSensor *nodeSensor = new SoNodeSensor(sceneChangeCB, &gv);
      nodeSensor->attach(gv.scene);
    }
    
    SoXt::show(myWindow); // Display main window
    SoXt::mainLoop();	// Main Inventor event loop
  }

  // Or Render and Quit (no display)
  else {

    if (gv.algorithm != 2 && gv.algorithm !=3 && gv.inImgFile.empty() && 
	( (gv.renderFlags == 0 || gv.renderFlags & Pane::C_CAUSTIC) &&
	  gv.numCPhotons>0) ||
	( (gv.algorithm==1 || gv.renderFlags & Pane::C_MDIFFUSE)    &&
	  gv.numPhotons >0) ||
	gv.numVPhotons > 0 )
      pane.shootPhotons();

    // render a series of frames
    if (gv.animateScene){

      // check camera type and number of cameras
      if (! gv.cam->isOfType(SoPerspectiveCamera::getClassTypeId()) ){
	cerr << "Error: Scene camera is not a SoPerspectiveCamera, cannot animate.\n";
	exit(1);
      }
      if (gv.camList->numCams()<=0){
	cerr << "Error: No cameras defined in camera list (camlist.iv), cannot animate.\n";
	exit(2);
      }

      float time;
      cerr << "Rendering " << gv.frames << " frames.\n";

      // start at frame determined by time
      int startFrame = int(gv.timeVal*(gv.frames-1));
      int endFrame = gv.frames-1;

      // configure a range of frames if --framessegment specified
      if (gv.framesSegment[1]>0 && gv.framesSegment[1]<=gv.frames){
	float t[2];
	t[0] = (gv.framesSegment[0]-1)*1.0/gv.framesSegment[1];
	t[1] = (gv.framesSegment[0]  )*1.0/gv.framesSegment[1];
	startFrame = (int)(t[0] * (gv.frames-1));
	endFrame   = (int)(t[1] * (gv.frames-1));
	
	if (  (t[0] * (gv.frames-1)) - startFrame > 0.0)
	  startFrame++;

	if (  (t[1] * (gv.frames-1)) - endFrame <= 0.0)
	  endFrame--;
	
	if (gv.framesSegment[0] == gv.framesSegment[1])
	  endFrame = gv.frames-1;

	cerr << "Rendering frames " << startFrame << " through " << endFrame
	     << " (segment " << gv.framesSegment[0] << "/" << gv.framesSegment[1] << ").\n";
      }

      string origName = gv.imgFile;
      
      for (int i=startFrame; i<=endFrame; i++){
	gv.timeMin =                    i /(float)(gv.frames-1);
	gv.timeMax = MIN2(gv.frames-1,i+1)/(float)(gv.frames-1);
	time       = gv.timeMin;
	cerr << "Rendering frame " << i << ", " << i+1 << "/" << gv.frames << "\n";
	cerr << "Time = [" << gv.timeMin << ", " << gv.timeMax << "]" << endl;

	// position camera
	if (gv.cam->isOfType(SoPerspectiveCamera::getClassTypeId()))
	  gv.camList->positionCameraCubic( (SoPerspectiveCamera *) gv.cam,
					   (1.0-cos(time*M_PI))/2.0);

	// create filename
	char num[16] = {0};
	sprintf(num, "%04d", i);
        gv.imgFile = string("frame.") + num + "." + origName;
	
	// render
	pane.render();
	pane.writeImg();
      }

      // restore filename
      gv.imgFile = origName;
      
      // write last frame twice to make a nice series of file names
      //pane.writeImg();

    }
    
    // render IV file
    else if (gv.renderIv){
      pane.render();          // calls renderIv() after initial setup
      pane.writeIv();
    }

    // do image processing only
    else if (!gv.inImgFile.empty()){
      pane.writeImg();
    }
    
    // render one frame
    else {
      pane.render(gv.debugPixel[X]>=0 && gv.debugRange<0);
      pane.writeImg();
    }

    // erase scene graph (for memory debugging)
    gv.root->unref();

    // print out total rendering time

    // user time
    struct tms processTime;
    times(&processTime);

//   cerr << "utime = " << processTime.tms_utime << "\n";
//   cerr << "stime = " << processTime.tms_stime << "\n";
//   cerr << "cutime = " << processTime.tms_cutime << "\n";
//   cerr << "cstime = " << processTime.tms_cstime << "\n";
//   cerr << "totalTicks = " << gv.times.totalTicks << "\n";
//   cerr << "renderTicks = " << gv.times.renderTicks << "\n";
  
    gv.times.totalTicks += processTime.tms_utime + processTime.tms_cutime;
    clock_t totalTicks = gv.times.totalTicks;
    long ticksPerSec = sysconf(_SC_CLK_TCK);
    float ticksMs =  totalTicks/(double)ticksPerSec*1000.0;
    print ("\nTotal cpu ticks: %d, Ticks/s: %d\n", totalTicks, ticksPerSec);
    print("Total cpu time   : (%f ms, %f s, %f m, %f h)\n",
	  ticksMs, ticksMs/1000.0, ticksMs/1000.0/60.0,
	  ticksMs/1000.0/60.0/60.0);

    // actual time
    gettimeofday(&totalEnd, NULL);   // end delay measure
    int ms = delay(totalStart, totalEnd);
    print("Total actual time: (%d ms, %f s, %f m, %f h)\n",
	  ms, ms/1000.0, ms/1000.0/60.0,
	  ms/1000.0/60.0/60.0);
  }

  return(0);
}




// SunOS stuff

#ifdef SUNOS

char* strsep(char** str, const char* delims){
    char* token;

    if (*str==NULL) {
        /* No more tokens */
        return NULL;
    }

    token=*str;
    while (**str!='\0') {
        if (strchr(delims,**str)!=NULL) {
            **str='\0';
            (*str)++;
            return token;
        }
        (*str)++;
    }
    /* There is no other token */
    *str=NULL;
    return token;
}

#endif

//} /* namespace paneMain */
  

//'''''''''''''WW'''''
//                    
//WWWW W   WWW WW WWW 
//WW WW W W WW WW WW W
//WW WW W W WW WW WW W
//WW WW W  WWW WW WW W
//                    
//''''''''''''''''''''
//
// Entry point for pane executable. Calls progMain()
//
int main(int argc, char *argv[]){
  int retVal = progMain(argc, argv);
#ifdef DEBUG_MEM  
  DumpUnfreed();
#endif // DEBUG_MEM
  return(retVal);
}

  
