
#include "Sky.h"
#include "Spectrum.h"
#include "DataSet.h"

//     Y N
//     |/
//  W--x--X(E)
//    /|
//   Z(S)



// Data stolen from:
// http://www.cs.utah.edu/vissim/papers/sunsky/code/RiSunConstants.C

// 380-750 by 10nm
static float solAmplitudes[38] = {
  165.5, 162.3, 211.2, 258.8, 258.2,
  242.3, 267.6, 296.6, 305.4, 300.6,
  306.6, 288.3, 287.1, 278.2, 271.0,
  272.3, 263.6, 255.0, 250.6, 253.1,
  253.5, 251.3, 246.3, 241.7, 236.8,
  232.1, 228.2, 223.4, 219.7, 215.3,
  211.0, 207.3, 202.4, 198.7, 194.3,
  190.7, 186.3, 182.6
};

Sky::Sky(){}

void Sky::init(GlobalVars *gv_) {

  // set GlobalVars pointer
  gv = gv_;
  
  float stdTime = 12+3;     // hours, decimal. 3pm
  float julianDate = 322;   // 11/18/07  see: `date +%j`
  float stdMeridianPST = -8.*15 * M_PI/180;  // radians
  float stdMeridian = stdMeridianPST;
  float latitude = 34.017 * M_PI/180;  // los angeles
  float longitude = -118.41 * M_PI/180; // http://www.gorissen.info/Pierre/maps/googleMapLocationv3.php

  // in decimal hours
  float solarTime = stdTime +
    0.170 * sin(4.*M_PI*(julianDate-80)/373.) -
    0.129 * sin(2.*M_PI*(julianDate-8)/355.) +
    12. * (stdMeridian-longitude)/M_PI;

  float solarDeclination = 0.4093 * sin(2.*M_PI*(julianDate-81)/368.); // radians "delta"
  float tmp = M_PI*solarTime/12.;

  // solar angle from zenith in radians
  // [0, pi/2]
  float thetaSolar = M_PI/2. - asin(sin(latitude)*sin(solarDeclination) -
				    cos(latitude)*cos(solarDeclination) *
				    cos(tmp));

  // solar azimuth in radians, west from South
  // [0, 2pi]
  float phiSolar = atan(-cos(solarDeclination)*sin(tmp) /
			(cos(latitude)*sin(solarDeclination) -
			 sin(latitude)*cos(solarDeclination)*cos(tmp)));

  // sun distance, in meters. perihelion/aphelion
  // distances from http://www.physicalgeography.net/fundamentals/6h.html
  float sunDistJul4 = 152.1e9; // far
  float sunDistJan3 = 147.3e9; // near

  // lerp... not accurate but whateva
  float yearFrac = (julianDate-3+365%365)/365.;
  if (yearFrac>.5) yearFrac = 1-yearFrac;
  float sunDist = (1-yearFrac)*sunDistJan3 + yearFrac*sunDistJul4;

  float sunRadius = 695500e3; // in meters (thanks google)
  sunRadiusAngle = atan2(sunRadius, sunDist);  //sunSize/sunDist;

  // orthonormal basis pointed towards sun
  float phi = 1.5*M_PI-phiSolar;
  sunBasisW.setValue(sin(thetaSolar)*cos(phi),
		     cos(thetaSolar),
		     -sin(thetaSolar)*sin(phi));
  sunBasisV.setValue(sunBasisW[1], sunBasisW[2], sunBasisW[0]);
  sunBasisV -= sunBasisV.dot(sunBasisW)*sunBasisW;
  sunBasisV.normalize();
  sunBasisU = sunBasisV.cross(sunBasisW);

  cerr << "Sun W = " << sunBasisW << endl;
  // sun color
  sunRadiance = Spectrum(solAmplitudes, 38, 380, 750).getRGB();
  sunRadiance *= 10 * 100*100;
  cerr << "Sun radiance = " << sunRadiance << endl;
  //exit(1);
}

SbColor Sky::impSample(Ray &r, SbVec3f &n, SbVec3f &ng, int simple) {

  int samples = 25;
  int sqrt_samples = (int)sqrt((float)samples);
  if (sqrt_samples<1) sqrt_samples = 1;
  int numVis = 0;
  SbColor color(0,0,0);
  for (int sampy=0; sampy<sqrt_samples; sampy++){
    for (int sampx=0; sampx<sqrt_samples; sampx++){

      // jittered
      double eps1 = (sampx + r.ctx->mydrand48()) / sqrt_samples;
      double eps2 = (sampy + r.ctx->mydrand48()) / sqrt_samples;

      double cosAlpha = 1+eps1*(cos(sunRadiusAngle)-1);
      double sinAlpha = sqrt(1-cosAlpha*cosAlpha);
      double phi = 2*M_PI*eps2;
      SbVec3f dir = 
 	sinAlpha*cos(phi)*sunBasisU +
 	sinAlpha*sin(phi)*sunBasisV +
 	cosAlpha*sunBasisW;

      // dont illuminate lights below horizon
      if (ng.dot(dir) <= 0.0)
	continue;
      
      // don't illuminate area that think they are facing away
      float dot1 = n.dot(dir);
      if (dot1<=0.0)
	continue;      
		  
      const float t_eps = 0.002;
      Ray lr = Ray(r.x+dir*t_eps, dir, r.ctx);
      lr.ignore(r.triId);
      if (gv->pane->rayAccel->intersect_ray(lr))
	continue; // hits something (and not background)

      float value = dot1 / (sqrt_samples*sqrt_samples);

      // ray marching absorbtion
      if (gv->stepSize>0){
	float tau = 0.0;
	float sdt = gv->stepSize;
	float sigma_t = 0.0;
	SbVec3f lx = r.x+gv->pane->getMaxDist()*dir;
	tau = gv->pane->dataSet->bresenSum(r.x, lx);
	tau = gv->meters(tau);
	value *= exp(-tau);
      }
	  
      // compute color
      // shirley p.120 eq. 15.5
      color += SbColor(r.diffCol[R] * M_1_PI * sunRadiance[R] * value,
		       r.diffCol[G] * M_1_PI * sunRadiance[G] * value,
		       r.diffCol[B] * M_1_PI * sunRadiance[B] * value) ;

    }
  }

  return color;
}
float Sky::getSample(double eps1, double eps2,
			  SbVec3f &dir, SbColor &c){
  return 0;
}
SbColor Sky::getRadiance(SbVec3f &dir){

  if (acos(dir.dot(sunBasisW))<sunRadiusAngle)
    return sunRadiance;
  
  return SbColor(0,0,0);
}
